@@ -2,7 +2,7 @@ package Couchbase::Client;
 
 BEGIN {
     require XSLoader;
-    our $VERSION = '0.18';
+    our $VERSION = '1.0.0';
     XSLoader::load(__PACKAGE__, $VERSION);
 }
 
@@ -197,22 +197,6 @@ see L<Couchbase::Client::README> for a broader overview.
 
 See that same page for a list of current known issues as well.
 
-=head1 WARNING
-
-The bundled C<libcouchbase> is not an official release version, and might break.
-
-Until this module is bundled with an official release version, assume any bug is
-a result of the perl bindings, and/or my modifications to the library, and not
-something in C<libcouchbase> itself.
-
-This warning will probably be removed in a stable version.
-
-The only reason this module is on CPAN is to get smoke tests.
-
-That being said, the module has been quite stable for me, and should be offered
-in a non-underscore release in the near future.
-
-
 =head1 SYNOPSIS
 
     use Couchbase::Client;
