@@ -23,6 +23,8 @@
  */
 #include "internal.h"
 
+/* private function to safely free backup_nodes*/
+static void free_backup_nodes(libcouchbase_t instance);
 /**
  * Get the version of the library.
  *
@@ -55,6 +57,74 @@ const char *libcouchbase_get_port(libcouchbase_t instance)
     return instance->port;
 }
 
+static void setup_current_host(libcouchbase_t instance, const char *host)
+{
+    char *ptr;
+    snprintf(instance->host, sizeof(instance->host), "%s", host);
+    if ((ptr = strchr(instance->host, ':')) == NULL) {
+        strcpy(instance->port, "8091");
+    } else {
+        *ptr = '\0';
+        snprintf(instance->port, sizeof(instance->port), "%s", ptr + 1);
+    }
+}
+
+static int setup_boostrap_hosts(libcouchbase_t ret, const char *host)
+{
+    const char *ptr = host;
+    int num = 0;
+    int ii;
+
+    while ((ptr = strchr(ptr, ';')) != NULL) {
+        ++ptr;
+        ++num;
+    }
+
+    /* Let's allocate the buffer space and copy the pointers */
+    if ((ret->backup_nodes = calloc(num + 2, sizeof(char *))) == NULL) {
+        return -1;
+    }
+
+    ret->nbackup_nodes = num + 2;
+    ret->should_free_backup_nodes = 1;
+
+    ptr = host;
+    ii = 0;
+    do {
+        char nm[NI_MAXHOST + NI_MAXSERV + 2];
+        const char *start = ptr;
+        ptr = strchr(ptr, ';');
+        if (ptr == NULL) {
+            /* this is the last part */
+            ret->backup_nodes[ii] = strdup(start);
+            ptr = NULL;
+        } else {
+            /* copy everything up to ';' */
+            unsigned long size = (unsigned long)ptr - (unsigned long)start;
+            if (size < sizeof(nm)) {
+                memcpy(nm, start, ptr - start);
+                *(nm + size) = '\0';
+            } else {
+                /* skipping the hostname because it's too long */
+            }
+            ++ptr;
+            ret->backup_nodes[ii] = strdup(nm);
+        }
+        if (ret->backup_nodes[ii] == NULL) {
+            do {
+                free(ret->backup_nodes[ii--]);
+            } while (ii > -1);
+            return -1;
+        }
+        ++ii;
+    } while (ptr != NULL);
+
+    ret->backup_idx = 0;
+    setup_current_host(ret, ret->backup_nodes[0]);
+
+    return 0;
+}
+
 LIBCOUCHBASE_API
 libcouchbase_t libcouchbase_create(const char *host,
                                    const char *user,
@@ -65,7 +135,6 @@ libcouchbase_t libcouchbase_create(const char *host,
     char buffer[1024];
     libcouchbase_ssize_t offset;
     libcouchbase_t ret;
-    char *p;
 
     if (io == NULL) {
         io = libcouchbase_create_io_ops(LIBCOUCHBASE_IO_OPS_DEFAULT,
@@ -94,12 +163,9 @@ libcouchbase_t libcouchbase_create(const char *host,
     libcouchbase_initialize_packet_handlers(ret);
     libcouchbase_behavior_set_syncmode(ret, LIBCOUCHBASE_ASYNCHRONOUS);
 
-    ret->host = strdup(host);
-    if ((p = strchr(ret->host, ':')) == NULL) {
-        ret->port = "8091";
-    } else {
-        *p = '\0';
-        ret->port = p + 1;
+    if (setup_boostrap_hosts(ret, host) == -1) {
+        free(ret);
+        return NULL;
     }
 
     offset = snprintf(buffer, sizeof(buffer),
@@ -118,10 +184,10 @@ libcouchbase_t libcouchbase_create(const char *host,
         offset += snprintf(buffer + offset, sizeof(buffer) - (libcouchbase_size_t)offset,
                            "Authorization: Basic %s\r\n", base64);
     }
-    offset += snprintf(buffer + offset, sizeof(buffer)-(libcouchbase_size_t)offset, "\r\n");
+    offset += snprintf(buffer + offset, sizeof(buffer) - (libcouchbase_size_t)offset, "\r\n");
     ret->http_uri = strdup(buffer);
 
-    if (ret->host == NULL || ret->http_uri == NULL) {
+    if (ret->http_uri == NULL) {
         libcouchbase_destroy(ret);
         return NULL;
     }
@@ -145,7 +211,6 @@ LIBCOUCHBASE_API
 void libcouchbase_destroy(libcouchbase_t instance)
 {
     libcouchbase_size_t ii;
-    free(instance->host);
     free(instance->http_uri);
 
     if (instance->sock != INVALID_SOCKET) {
@@ -172,8 +237,9 @@ void libcouchbase_destroy(libcouchbase_t instance)
     for (ii = 0; ii < instance->nservers; ++ii) {
         libcouchbase_server_destroy(instance->servers + ii);
     }
+
+    free_backup_nodes(instance);
     free(instance->servers);
-    free(instance->backup_nodes);
 
     if (instance->io && instance->io->destructor) {
         instance->io->destructor(instance->io);
@@ -236,29 +302,35 @@ static int sasl_get_password(sasl_conn_t *conn, void *context, int id,
     return SASL_OK;
 }
 
-static void apply_vbucket_config(libcouchbase_t instance, VBUCKET_CONFIG_HANDLE config)
+libcouchbase_error_t libcouchbase_apply_vbucket_config(libcouchbase_t instance, VBUCKET_CONFIG_HANDLE config)
 {
     libcouchbase_uint16_t ii, max;
     libcouchbase_size_t num;
     const char *passwd;
     char curnode[NI_MAXHOST + NI_MAXSERV + 2];
     sasl_callback_t sasl_callbacks[4] = {
-        { SASL_CB_USER, (int(*)(void))&sasl_get_username, instance },
-        { SASL_CB_AUTHNAME, (int(*)(void))&sasl_get_username, instance },
-        { SASL_CB_PASS, (int(*)(void))&sasl_get_password, instance },
+        { SASL_CB_USER, (int( *)(void)) &sasl_get_username, instance },
+        { SASL_CB_AUTHNAME, (int( *)(void)) &sasl_get_username, instance },
+        { SASL_CB_PASS, (int( *)(void)) &sasl_get_password, instance },
         { SASL_CB_LIST_END, NULL, NULL }
     };
 
-    num = (libcouchbase_size_t)vbucket_config_get_num_servers(config);
-    instance->nservers = num;
-    instance->servers = calloc(num, sizeof(libcouchbase_server_t));
     instance->vbucket_config = config;
-    if (instance->backup_nodes) {
-        free(instance->backup_nodes);
+    num = (libcouchbase_size_t)vbucket_config_get_num_servers(config);
+    /* servers array should be freed in the caller */
+    instance->servers = calloc(num, sizeof(libcouchbase_server_t));
+    if (instance->servers == NULL) {
+        return libcouchbase_error_handler(instance, LIBCOUCHBASE_ENOMEM, "Failed to allocate memory");
     }
+    instance->nservers = num;
+    free_backup_nodes(instance);
     instance->backup_nodes = calloc(num, sizeof(char *));
-    sprintf(curnode, "%s:%s", instance->host, instance->port);
-    for (ii = 0; ii < (libcouchbase_size_t)num; ++ii) {
+    if (instance->backup_nodes == NULL) {
+        return libcouchbase_error_handler(instance, LIBCOUCHBASE_ENOMEM, "Failed to allocate memory");
+    }
+    instance->nbackup_nodes = num;
+    snprintf(curnode, sizeof(curnode), "%s:%s", instance->host, instance->port);
+    for (ii = 0; ii < num; ++ii) {
         instance->servers[ii].instance = instance;
         libcouchbase_server_initialize(instance->servers + ii, (int)ii);
         if (strcmp(curnode, instance->servers[ii].rest_api_server) == 0) {
@@ -280,7 +352,11 @@ static void apply_vbucket_config(libcouchbase_t instance, VBUCKET_CONFIG_HANDLE
     passwd = vbucket_config_get_password(instance->vbucket_config);
     if (passwd) {
         instance->sasl.password.secret.len = strlen(passwd);
-        strcpy((char*)instance->sasl.password.secret.data, passwd);
+        if (instance->sasl.password.secret.len < sizeof(instance->sasl.password.buffer) - offsetof(sasl_secret_t, data)) {
+            memcpy(instance->sasl.password.secret.data, passwd, instance->sasl.password.secret.len);
+        } else {
+            return libcouchbase_error_handler(instance, LIBCOUCHBASE_EINVAL, "Password too long");
+        }
     }
     memcpy(instance->sasl.callbacks, sasl_callbacks, sizeof(sasl_callbacks));
 
@@ -293,9 +369,13 @@ static void apply_vbucket_config(libcouchbase_t instance, VBUCKET_CONFIG_HANDLE
     instance->nvbuckets = max;
     free(instance->vb_server_map);
     instance->vb_server_map = calloc(max, sizeof(libcouchbase_vbucket_t));
+    if (instance->vb_server_map == NULL) {
+        return libcouchbase_error_handler(instance, LIBCOUCHBASE_ENOMEM, "Failed to allocate memory");
+    }
     for (ii = 0; ii < max; ++ii) {
         instance->vb_server_map[ii] = vbucket_get_master(instance->vbucket_config, ii);
     }
+    return LIBCOUCHBASE_SUCCESS;
 }
 
 static void relocate_packets(libcouchbase_server_t *src,
@@ -322,7 +402,11 @@ static void relocate_packets(libcouchbase_server_t *src,
         vb = ntohs(cmd.request.vbucket);
         idx = (libcouchbase_size_t)vbucket_get_master(dst_instance->vbucket_config, vb);
         dst = dst_instance->servers + idx;
-        assert(libcouchbase_ringbuffer_read(&src->output_cookies, &ct, sizeof(ct)) == sizeof(ct));
+        if (src->connected) {
+            assert(libcouchbase_ringbuffer_read(&src->output_cookies, &ct, sizeof(ct)) == sizeof(ct));
+        } else {
+            assert(libcouchbase_ringbuffer_read(&src->pending_cookies, &ct, sizeof(ct)) == sizeof(ct));
+        }
 
         assert(libcouchbase_ringbuffer_ensure_capacity(&dst->cmd_log, npacket));
         assert(libcouchbase_ringbuffer_write(&dst->cmd_log, cmd.bytes, sizeof(cmd.bytes)) == sizeof(cmd.bytes));
@@ -352,7 +436,7 @@ static void libcouchbase_update_serverlist(libcouchbase_t instance)
 {
     libcouchbase_size_t ii;
     VBUCKET_CONFIG_HANDLE next_config, curr_config;
-    VBUCKET_CONFIG_DIFF* diff = NULL;
+    VBUCKET_CONFIG_DIFF *diff = NULL;
     libcouchbase_size_t nservers;
     libcouchbase_server_t *servers, *ss;
 
@@ -367,6 +451,7 @@ static void libcouchbase_update_serverlist(libcouchbase_t instance)
                              instance->vbucket_stream.input.data) != 0) {
         libcouchbase_error_handler(instance, LIBCOUCHBASE_PROTOCOL_ERROR,
                                    vbucket_get_error_message(next_config));
+        vbucket_config_destroy(next_config);
         return;
     }
     instance->vbucket_stream.input.avail = 0;
@@ -377,7 +462,11 @@ static void libcouchbase_update_serverlist(libcouchbase_t instance)
             VBUCKET_DISTRIBUTION_TYPE dist_t = vbucket_config_get_distribution_type(next_config);
             nservers = instance->nservers;
             servers = instance->servers;
-            apply_vbucket_config(instance, next_config);
+            if (libcouchbase_apply_vbucket_config(instance, next_config) != LIBCOUCHBASE_SUCCESS) {
+                vbucket_free_diff(diff);
+                vbucket_config_destroy(next_config);
+                return;
+            }
             for (ii = 0; ii < nservers; ++ii) {
                 ss = servers + ii;
                 if (dist_t == VBUCKET_DISTRIBUTION_VBUCKET) {
@@ -390,6 +479,7 @@ static void libcouchbase_update_serverlist(libcouchbase_t instance)
                 }
                 libcouchbase_server_destroy(ss);
             }
+            free(servers);
 
             /* Destroy old config */
             vbucket_config_destroy(curr_config);
@@ -407,10 +497,16 @@ static void libcouchbase_update_serverlist(libcouchbase_t instance)
         } else {
             vbucket_config_destroy(next_config);
         }
+        if (diff) {
+            vbucket_free_diff(diff);
+        }
     } else {
         assert(instance->servers == NULL);
         assert(instance->nservers == 0);
-        apply_vbucket_config(instance, next_config);
+        if (libcouchbase_apply_vbucket_config(instance, next_config) != LIBCOUCHBASE_SUCCESS) {
+            vbucket_config_destroy(next_config);
+            return;
+        }
 
         /* Notify anyone interested in this event... */
         if (instance->vbucket_state_listener != NULL) {
@@ -432,7 +528,7 @@ static int parse_chunk(libcouchbase_t instance)
     buffer_t *buffer = &instance->vbucket_stream.chunk;
     assert(instance->vbucket_stream.chunk_size != 0);
 
-    if (instance->vbucket_stream.chunk_size == (libcouchbase_size_t)-1) {
+    if (instance->vbucket_stream.chunk_size == (libcouchbase_size_t) - 1) {
         char *ptr = strstr(buffer->data, "\r\n");
         long val;
         if (ptr == NULL) {
@@ -481,12 +577,12 @@ static int parse_header(libcouchbase_t instance)
     }
 
     /* parse the headers I care about... */
-    if(sscanf(buffer->data, "HTTP/1.1 %d", &response_code) != 1) {
+    if (sscanf(buffer->data, "HTTP/1.1 %d", &response_code) != 1) {
         libcouchbase_error_handler(instance, LIBCOUCHBASE_PROTOCOL_ERROR,
                                    buffer->data);
-    } else if(response_code != 200) {
+    } else if (response_code != 200) {
         libcouchbase_error_t err;
-        switch(response_code) {
+        switch (response_code) {
         case 401:
             err = LIBCOUCHBASE_AUTH_ERROR;
             break;
@@ -502,7 +598,7 @@ static int parse_header(libcouchbase_t instance)
     }
 
     if (strstr(buffer->data, "Transfer-Encoding: chunked") == NULL &&
-        strstr(buffer->data, "Transfer-encoding: chunked") == NULL) {
+            strstr(buffer->data, "Transfer-encoding: chunked") == NULL) {
         libcouchbase_error_handler(instance, LIBCOUCHBASE_PROTOCOL_ERROR,
                                    buffer->data);
         return -1;
@@ -513,7 +609,7 @@ static int parse_header(libcouchbase_t instance)
     buffer->avail -= (libcouchbase_size_t)(ptr - buffer->data);
     memmove(buffer->data, ptr, buffer->avail);
     buffer->data[buffer->avail] = '\0';
-    instance->vbucket_stream.chunk_size = (libcouchbase_size_t)-1;
+    instance->vbucket_stream.chunk_size = (libcouchbase_size_t) - 1;
 
     return 0;
 }
@@ -530,7 +626,8 @@ const libcouchbase_size_t min_buffer_size = 2048;
  * @param min_free the minimum amount of free space I need
  * @return 1 if success, 0 otherwise
  */
-int grow_buffer(buffer_t *buffer, libcouchbase_size_t min_free) {
+int grow_buffer(buffer_t *buffer, libcouchbase_size_t min_free)
+{
     if (min_free == 0) {
         /*
         ** no minimum size requested, just ensure that there is at least
@@ -575,41 +672,29 @@ static int libcouchbase_switch_to_backup_node(libcouchbase_t instance,
                                               libcouchbase_error_t error,
                                               const char *reason)
 {
-    libcouchbase_size_t nn;
-    char *pp;
-    int connected;
+    int connected = 0;
 
     if (instance->backup_nodes == NULL) {
         libcouchbase_error_handler(instance, error, reason);
         return -1;
     }
-    connected = 0;
-    nn = 0;
-    while (!connected) {
-        char *oldhost = instance->host;
-        libcouchbase_error_t rc;
 
-        while (instance->backup_nodes[nn] == NULL && nn < instance->nservers) {
-            nn++;
-        }
-        if (instance->backup_nodes[nn] == NULL) {
+    ++instance->backup_idx;
+    if (instance->backup_nodes[instance->backup_idx] == NULL) {
+        --instance->backup_idx;
+        libcouchbase_error_handler(instance, error, reason);
+        return -1;
+    }
+
+    while (!connected) {
+        /* Keep on trying the nodes until all of them failed ;-) */
+        libcouchbase_error_t rc = libcouchbase_connect(instance);
+        connected = (rc == LIBCOUCHBASE_SUCCESS);
+        if (!connected && instance->backup_nodes[instance->backup_idx] == NULL) {
             libcouchbase_error_handler(instance, LIBCOUCHBASE_NETWORK_ERROR,
                                        "failed to get config. All known nodes are dead.");
             return -1;
         }
-        instance->host = strdup(instance->backup_nodes[nn]);
-        instance->backup_nodes[nn] = NULL;
-        if ((pp = strchr(instance->host, ':')) == NULL) {
-            instance->port = "8091";
-        } else {
-            *pp = '\0';
-            instance->port = pp + 1;
-        }
-        free(oldhost);
-
-        /* try to connect to next node */
-        rc = libcouchbase_connect(instance);
-        connected = (rc == LIBCOUCHBASE_SUCCESS);
     }
     return 0;
 }
@@ -618,7 +703,7 @@ static void libcouchbase_instance_connerr(libcouchbase_t instance,
                                           libcouchbase_error_t err,
                                           const char *errinfo)
 {
-    if(instance->sock != INVALID_SOCKET) {
+    if (instance->sock != INVALID_SOCKET) {
         instance->io->delete_event(instance->io, instance->sock, instance->event);
         instance->io->close(instance->io, instance->sock);
         instance->sock = INVALID_SOCKET;
@@ -767,7 +852,7 @@ static void vbucket_stream_handler(libcouchbase_socket_t sock, short which, void
                     libcouchbase_update_serverlist(instance);
                 }
 
-                instance->vbucket_stream.chunk_size = (libcouchbase_size_t)-1;
+                instance->vbucket_stream.chunk_size = (libcouchbase_size_t) - 1;
                 if (buffer->avail > 0) {
                     done = 0;
                 }
@@ -781,6 +866,7 @@ static void vbucket_stream_handler(libcouchbase_socket_t sock, short which, void
 
 static void libcouchbase_instance_connected(libcouchbase_t instance)
 {
+    instance->backup_idx = 0;
     instance->io->update_event(instance->io, instance->sock,
                                instance->event, LIBCOUCHBASE_RW_EVENT,
                                instance, vbucket_stream_handler);
@@ -793,20 +879,14 @@ static void libcouchbase_instance_connect_handler(libcouchbase_socket_t sock,
     libcouchbase_t instance = arg;
     int retry;
     int first_try = (sock == INVALID_SOCKET);
-
+    libcouchbase_connect_status_t connstatus = LIBCOUCHBASE_CONNECT_OK;
+    int save_errno;
     do {
         if (instance->sock == INVALID_SOCKET) {
             /* Try to get a socket.. */
-            while (instance->curr_ai != NULL) {
-                instance->sock = instance->io->socket(instance->io,
-                                                      instance->curr_ai->ai_family,
-                                                      instance->curr_ai->ai_socktype,
-                                                      instance->curr_ai->ai_protocol);
-                if (instance->sock != INVALID_SOCKET) {
-                    break;
-                }
-                instance->curr_ai = instance->curr_ai->ai_next;
-            }
+            instance->sock = libcouchbase_gai2sock(instance,
+                                                   &instance->curr_ai,
+                                                   &save_errno);
 
             /* Reset the stream state, we run this only during a new socket. */
             libcouchbase_instance_reset_stream_state(instance);
@@ -814,17 +894,23 @@ static void libcouchbase_instance_connect_handler(libcouchbase_socket_t sock,
 
         if (instance->curr_ai == NULL) {
             char errinfo[1024];
-            snprintf(errinfo, sizeof(errinfo), "Failed to look up \"%s:%s\"",
-                     instance->host, instance->port);
+            libcouchbase_error_t our_errno;
+            libcouchbase_sockconn_errinfo(save_errno,
+                                          instance->host,
+                                          instance->port,
+                                          instance->ai,
+                                          errinfo,
+                                          sizeof(errinfo),
+                                          &our_errno);
+
             if (first_try && instance->sock != INVALID_SOCKET) {
                 /* Ensure our connerr function doesn't try to delete a
                  * nonexistent event */
                 instance->io->close(instance->io, instance->sock);
                 instance->sock = INVALID_SOCKET;
             }
-            libcouchbase_instance_connerr(instance,
-                                          LIBCOUCHBASE_NETWORK_ERROR,
-                                          errinfo);
+
+            libcouchbase_instance_connerr(instance, our_errno, errinfo);
             return ;
         }
 
@@ -836,15 +922,17 @@ static void libcouchbase_instance_connect_handler(libcouchbase_socket_t sock,
             libcouchbase_instance_connected(instance);
             return ;
         } else {
-            switch (instance->io->error) {
-            case EINTR:
+            save_errno = instance->io->error;
+            connstatus = libcouchbase_connect_status(save_errno);
+
+            switch (connstatus) {
+            case LIBCOUCHBASE_CONNECT_EINTR:
                 retry = 1;
                 break;
-            case EISCONN:
+            case LIBCOUCHBASE_CONNECT_EISCONN:
                 libcouchbase_instance_connected(instance);
                 return ;
-            case EWOULDBLOCK:
-            case EINPROGRESS: /* First call to connect */
+            case LIBCOUCHBASE_CONNECT_EINPROGRESS:
                 instance->io->update_event(instance->io,
                                            instance->sock,
                                            instance->event,
@@ -852,13 +940,11 @@ static void libcouchbase_instance_connect_handler(libcouchbase_socket_t sock,
                                            instance,
                                            libcouchbase_instance_connect_handler);
                 return ;
-            case EALREADY: /* Subsequent calls to connect */
+            case LIBCOUCHBASE_CONNECT_EALREADY: /* Subsequent calls to connect */
                 return ;
 
             default: {
                 /* Save errno because of possible subsequent errors */
-                int save_errno = instance->io->error;
-
                 if (instance->sock != INVALID_SOCKET) {
                     if (!first_try) {
                         /* Event updated */
@@ -869,19 +955,24 @@ static void libcouchbase_instance_connect_handler(libcouchbase_socket_t sock,
                     instance->io->close(instance->io, instance->sock);
                     instance->sock = INVALID_SOCKET;
                 }
-
-                if (save_errno != ECONNREFUSED) {
+                if (connstatus == LIBCOUCHBASE_CONNECT_EFAIL &&
+                        instance->curr_ai->ai_next) {
+                    /* Here we handle 'medium-type' errors which are not a hard
+                     * failure, but mean that we need to retry the connect() with
+                     * different parameters.
+                     */
+                    retry = 1;
+                    instance->curr_ai = instance->curr_ai->ai_next;
+                    break;
+                } else {
                     char errinfo[1024];
                     snprintf(errinfo, sizeof(errinfo), "Connection failed: %s",
                              strerror(instance->io->error));
                     libcouchbase_instance_connerr(instance,
-                                                  LIBCOUCHBASE_NETWORK_ERROR,
+                                                  LIBCOUCHBASE_CONNECT_ERROR,
                                                   errinfo);
                     return ;
                 }
-
-                retry = 1;
-                instance->curr_ai = instance->curr_ai->ai_next;
             }
 
             }
@@ -915,14 +1006,29 @@ libcouchbase_error_t libcouchbase_connect(libcouchbase_t instance)
     hints.ai_socktype = SOCK_STREAM;
     hints.ai_family = AF_UNSPEC;
 
-    error = getaddrinfo(instance->host, instance->port, &hints, &instance->ai);
-    if (error != 0) {
-        char errinfo[1024];
-        snprintf(errinfo, sizeof(errinfo), "Failed to look up \"%s:%s\"",
-                 instance->host, instance->port);
-        return libcouchbase_error_handler(instance, LIBCOUCHBASE_UNKNOWN_HOST,
-                                          errinfo);
-    }
+    do {
+        setup_current_host(instance,
+                           instance->backup_nodes[instance->backup_idx]);
+        error = getaddrinfo(instance->host, instance->port,
+                            &hints, &instance->ai);
+        if (error != 0) {
+            /* Ok, we failed to look up that server.. look up the next
+             * in the list
+             */
+            instance->backup_idx++;
+            if (instance->backup_nodes[instance->backup_idx] == NULL) {
+                char errinfo[1024];
+                snprintf(errinfo, sizeof(errinfo),
+                         "Failed to look up \"%s:%s\"",
+                         instance->host, instance->port);
+                return libcouchbase_error_handler(instance,
+                                                  LIBCOUCHBASE_UNKNOWN_HOST,
+                                                  errinfo);
+            }
+            setup_current_host(instance,
+                               instance->backup_nodes[instance->backup_idx]);
+        }
+    } while (error != 0);
 
     instance->curr_ai = instance->ai;
     instance->event = instance->io->create_event(instance->io);
@@ -931,3 +1037,16 @@ libcouchbase_error_t libcouchbase_connect(libcouchbase_t instance)
 
     return instance->last_error;
 }
+
+static void free_backup_nodes(libcouchbase_t instance)
+{
+    if (instance->should_free_backup_nodes) {
+        libcouchbase_size_t ii;
+        for (ii = 0; ii < instance->nbackup_nodes; ++ii) {
+            free(instance->backup_nodes[ii]);
+        }
+        instance->should_free_backup_nodes = 0;
+    }
+
+    free(instance->backup_nodes);
+}
