@@ -18,9 +18,11 @@
 #ifndef CONFIGURATION_H
 #define CONFIGURATION_H 1
 
+#include "config.h"
 #include <string>
 
-class Configuration {
+class Configuration
+{
 public:
     Configuration();
     ~Configuration();
@@ -34,6 +36,9 @@ public:
     const char *getBucket() const;
     void setTimingsEnabled(bool n);
     bool isTimingsEnabled() const;
+    void setTimeout(const char *t);
+    void setTimeout(uint32_t t);
+    uint32_t getTimeout(void);
 
 protected:
     void loadCbcRc(void);
@@ -43,6 +48,7 @@ protected:
     std::string passwd;
     std::string bucket;
     bool timings;
+    uint32_t timeout;
 };
 
 #endif
